/*                                                              -*-c++-*-
 * Copyright (c) 2015-2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Sep 16 2015
 */
#ifndef H_POM_CODELS
#define H_POM_CODELS

#include <aio.h>

#include <map>

#include "pom_c_types.h"
#include "ukf.h"

struct pom_snapshot {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

  bool fused;

  ukf::state_s state;
  ukf::state_s::dcov_s cov;

  or_pose_estimator_state measurement;
};

inline bool operator <(const or_time_ts &x, const or_time_ts &y)
{
  /* sorted by greater timestamps first */
  if (x.sec > y.sec) return true;
  if (x.sec == y.sec)
    if (x.nsec > y.nsec) return true;
  return false;
}

typedef std::multimap<
  or_time_ts, pom_snapshot,
  std::less<or_time_ts>,
  Eigen::aligned_allocator<
    std::pair<const or_time_ts, pom_snapshot> > > pom_history_s;

struct pom_context_s {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

  or_time_ts ts;
  ukf::state_s state;
  ukf::state_s::dcov_s cov;

  ukf::filter_s filter;
  pom_history_s history;
};

struct pom_log_s {
  struct aiocb req;
  char buffer[4096];
  bool pending, skipped;
  uint32_t decimation;
  size_t missed, total;

# define pom_logfmt	" %g"
# define pom_log_header_fmt                                             \
  "ts i posp attp velp avelp accp "                                     \
  "x y z roll pitch yaw vx vy vz wx wy wz ax ay az"
# define pom_log_state_fmt                                              \
  "%d.%09d %d %d %d %d %d %d"                                           \
  pom_logfmt pom_logfmt pom_logfmt pom_logfmt pom_logfmt pom_logfmt     \
  pom_logfmt pom_logfmt pom_logfmt pom_logfmt pom_logfmt pom_logfmt     \
  pom_logfmt pom_logfmt pom_logfmt
# define pom_scan_state_fmt                                             \
  "%d.%09d %d %d %d %d %d %d"                                           \
  "%lf %lf %lf %lf %lf %lf"                                             \
  "%lf %lf %lf %lf %lf %lf"                                             \
  "%lf %lf %lf"
};


static inline genom_event
pom_e_sys_error(const char *s, genom_context self)
{
  pom_e_sys_detail d;
  char buf[64], *p;

  d.code = errno;

#ifdef STRERROR_R_CHAR_P
  /* glibc managed to mess up with this function */
  p = strerror_r(d.code, buf, sizeof(buf));
#else
  strerror_r(d.code, buf, sizeof(buf));
  p = buf;
#endif
  snprintf(d.what, sizeof(d.what), "%s%s%s", s ? s : "", s ? ": " : "", p);

  return pom_e_sys(&d, self);
}

#endif /* H_POM_CODELS */
